//Saving results obtained
//in NONCOLAF.CPP
//(c)V.Glazkov, March 2016

#include"energy.h"
#include<string.h>
#include"cubic.h"
#include"nrutil.h"
#include"nr.h"

#define _cutoff -1e-4 //allowance for numeric uncertainty of omega^2

//global variables
extern vector3d dirH;
extern double Hval,Hstart,Hstop,Hstep,I[4],euler0[4];
extern int minsearch_flag,minsearch_grid;

//Functions declarations
//create file headers
void create_headers(char descrname[], char staticfilename[],char eigenfreqsfilename[],char eigenvectorsfilename[]);
//calculate and save static properties at given orientation
void save_static(char filename[], double theta, double phi, double psi);
//calculate and save eigenfrequencies of small oscillations
//at the vicinity of the specified equilibrium position
void save_eigenfreqs(char filename[], double theta, double phi, double psi, double * omegaout);
//calculate and save eigenvectors of small oscillations
//at the vicinity of the specified equilibrium position
void save_eigenvecs(char filename[], double theta, double phi, double psi, double omegain[]);


//Functions descriptions
#include<stdio.h>
char projection(int alpha)
	{
	char ch_alpha;
	switch(alpha)
		{
		case 1: ch_alpha='x';break;
		case 2: ch_alpha='y';break;
		case 3: ch_alpha='z';break;
		default: ch_alpha='?';break;
		}
	return (ch_alpha);
	}

void create_headers(char descrname[],char staticfilename[],char eigenfreqsfilename[],char eigenvectorsfilename[])
	{
	FILE * handle;
	int i,j,alpha,beta;

	handle=fopen(descrname,"w");

	fprintf(handle,"This file contains detailed description of the\n model parameters used by NONCOLAF program\nEnergy units is, normally, 10^9 1/sec (corresponds to GHz frequency range), field units is user choice (should be the same for H and gamma)\n");
	fprintf(handle,"\nApplied field direction: (%.4f,%.4f,%.4f)\n",(float)dirH.x,(float)dirH.y,(float)dirH.z);
	fprintf(handle,"Field scan: from %.4g to %.4g with step of %.4g\n",(float)Hstart,(float)Hstop,(float)Hstep);

	fprintf(handle,"\nAlgorithm settings:\n");
	if(minsearch_flag)
		fprintf(handle,"\tGlobal search for minimum over grid of %i points per euler angle\n",minsearch_grid);
	else
		fprintf(handle,"\tUse previous minimum (follow local minimum), starting approximation: theta=%.4g, phi=%.4g, psi=%.4g\n",(float)euler0[1],(float)euler0[2],(float)euler0[3]);

	fprintf(handle,"\nModel settings:\n");
	fprintf(handle,"\tGyromagnetic ratio:%.6g 10^9 sec^-1/field unit [corresponds to %.6g GHz/field unit]\n",(float)gamma,(float)(gamma/(2.*M_PI)));
	fprintf(handle,"\tI_1=%.6g, I_2=%.6g, I_3=%.6g\n",(float)I[1],(float)I[2],(float)I[3]);
	fprintf(handle,"\tI_i constants corresponds to following susceptibilities\nfor the field applied along i-th l-vector:\nchi_1=%.4g, chi_2=%.4g, chi_3=%.4g\n",(float)(gamma*gamma*(I[2]+I[3])),(float)(gamma*gamma*(I[3]+I[1])),(float)(gamma*gamma*(I[2]+I[1])));
	fprintf(handle,"\tAnisotropy energy:\n\tU_A=");
	for(i=1;i<=3;i++)
		for(j=1;j<=3;j++)
			for(alpha=1;alpha<=3;alpha++)
				for(beta=1;beta<=3;beta++)
					if(is_a[i][alpha][j][beta])
						{
						if(a[i][alpha][j][beta]>0.)
							fprintf(handle,"+");
						fprintf(handle,"%.6g*l%i%c*l%i%c",(float)a[i][alpha][j][beta],i,projection(alpha),j,projection(beta));
						}
	fprintf(handle,"\n");

	fclose(handle);
	handle=fopen(staticfilename,"w");
	fprintf(handle,"Static equilibrium properties of noncollinear antiferromagnet.\nModeled by NONCOLAF program.\nH\tE\tchi_par\tchi_perp\tTheta\tPhi\tPsi\t cos(l1,H)\t cos(l2,H)\t cos(l3,H)\n");
	fclose(handle);

	handle=fopen(eigenfreqsfilename,"w");
	fprintf(handle,"Oscillation eigenfrequencies of noncollinear antiferromagnet.\nModeled by NONCOLAF program.\nH\tf1\t,f2\tf3\n");
	fclose(handle);

	handle=fopen(eigenvectorsfilename,"w");
	fprintf(handle,"Oscillating magnetization components\nOscillating magnetization is complex vector,\neach component is shown as [Re+i*(Im)];\n");
	fprintf(handle,"\tAverage longitudinal and transverse projections\t\t\t\t Oscillation magnetization vectors\nH\tm1per\tm1par\tm2per\tm2par\tm3per\tm3par\t[m1_x];[m1_y];[m1_z];\t[m2_x];[m2_y];[m2_z];\t[m3_x];[m3_y];[m3_z];\n");
	fclose(handle);
	}

void save_static(char filename[],double theta, double phi, double psi)
	{
	double E, chipar,chiper, Hcos[4];
	FILE * handle;
	vector3d l[4];
	int i,code;

	for (i=1;i<=3;i++)
		{
		l[i]=l_vec(i,theta,phi,psi,&code);
		Hcos[i]=scalar(l[i],dirH);
		}

	E=U(l);
	susceptibilities(l,&chipar,&chiper);

	handle=fopen(filename,"a");
	fprintf(handle,"%.6f\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\n",(float)Hval,(float)E,(float)chipar,(float)chiper,(float)theta,(float)phi,(float)psi,(float)Hcos[1],(float)Hcos[2],(float)Hcos[3]);
	fclose(handle);
	}

void omega_matrices(void * omega2,void * iomega, double theta,double phi, double psi, int * err)
	{
	//omega2 and iomega matrices for eigenvectors and eigenfrequencies calculations
	vector3d l[4],dldphi[4][4],H;
	double * omega2pointer;
	double * iomegapointer;
	int i,alpha,beta,code;
	double _omega2[4][4],_iomega[4][4];

	H=Hval*dirH;

	omega2pointer=(double *)omega2;
	iomegapointer=(double *)iomega;

	//Clear arrays
	memset(_omega2,0,sizeof(_omega2));
	memset(_iomega,0,sizeof(_iomega));

	for(i=1;i<=3;i++)
		{
		l[i]=l_vec(i,theta,phi,psi,&code);
		for(alpha=1;alpha<=3;alpha++)
			dldphi[i][alpha]=dl_dphi(alpha,l[i],&code);
		}
	for(alpha=1;alpha<=3;alpha++)
		for(beta=alpha;beta<=3;beta++)
			{
			_omega2[alpha][beta]=_iomega[alpha][beta]=0.;
			for(i=1;i<=3;i++)
				{
				_omega2[alpha][beta]+=I[i]*scalar(dldphi[i][alpha],dldphi[i][beta]);
				if(alpha!=beta)
					_iomega[alpha][beta]+=2.*I[i]*gamma*scalar(dldphi[i][alpha],dldphi[i][beta]*H);
				}
			_omega2[beta][alpha]=_omega2[alpha][beta];
			_iomega[beta][alpha]=-_iomega[alpha][beta];
			}
	*err=code;
	for(alpha=0;alpha<=3;alpha++)
		for(beta=0;beta<=3;beta++)
			{
			*(omega2pointer+4*alpha+beta)=_omega2[alpha][beta];
			*(iomegapointer+4*alpha+beta)=_iomega[alpha][beta];
			}
	}
double _e(int i,int j, int k)
	{
	//Levi-Cavita e_ijk for determinant computation
	//e_ijk=0 if some of indices coinside
	//e_ijk=1 for 123 and ciclic permutations
	//e_ijk=-1 for 213 and ciclic permutations
	//i,j,k=1...3
	if(i<1||i>3||j<1||j>3||k<1||k>3)
		return(0.);
	if(i==j||j==k||k==i)
		return(0.);
	if((i==1&&j==2)||(i==2&&j==3)||(i==3&&j==1))
		return(1.);
	return(-1.);
	}
int _eint(int i,int j, int k)
	{
	//Levi-Cavita e_ijk for determinant computation
	//e_ijk=0 if some of indices coinside
	//e_ijk=1 for 123 and ciclic permutations
	//e_ijk=-1 for 213 and ciclic permutations
	//i,j,k=1...3
	if(i<1||i>3||j<1||j>3||k<1||k>3)
		return(0);
	if(i==j||j==k||k==i)
		return(0);
	if((i==1&&j==2)||(i==2&&j==3)||(i==3&&j==1))
		return(1);
	return(-1);
	}
void save_eigenfreqs(char filename[], double theta, double phi, double psi, double*omegaout)
	{
	//eigenfrequencies are calculated from
	//det(M)=0 equation
	//M[i][j]=-omega^2*_omega2[i][j]+i*omega*_iomega[i][j]+_hess[i][j]
	double _hess[4][4],_omega2[4][4],_iomega[4][4];
	int i,j,k,alpha,beta,code;
	double a6,a4,a2,a0,sorttmp;
	double omega[3],f[3];
	FILE * handle;


	//calculate hessian matrix
	Hessian(_hess,theta,phi,psi,&code);
	//calculate _omega2 and _iomega matrices
	omega_matrices(_omega2,_iomega,theta,phi,psi,&code);

	//det(M)=0 results in real numbers qubic equation
	//a6*omega^6+a4*omega^4+a2*omega^2+a0=0

	//calculating coefficients
	//note that _iomega matrix is multiplied by imaginary i=sqrt(-1)
	//and_omega2 matrix has a -1 multiplier

	a6=a4=a2=a0=0.;

	for(i=1;i<=3;i++)
		for(j=1;j<=3;j++)
			for(k=1;k<=3;k++)
				{
				if(_eint(i,j,k)==0) continue;
				//omega^6 appears only when _omega2[i][j] enter thrice, (-1)^3=-1 hence minus
				a6-=_e(i,j,k)*_omega2[1][i]*_omega2[2][j]*_omega2[3][k];
				//omega^4 appears when {_omega2 enter twice and _hess enter once, (-1)^2=1} or {_omega 2 enter once and _iomega enter twice -(i)^2=1}
				a4+=_e(i,j,k)*(_omega2[1][i]*_omega2[2][j]*_hess[3][k]+_omega2[1][i]*_hess[2][j]*_omega2[3][k]+_hess[1][i]*_omega2[2][j]*_omega2[3][k]+_omega2[1][i]*_iomega[2][j]*_iomega[3][k]+_iomega[1][i]*_omega2[2][j]*_iomega[3][k]+_iomega[1][i]*_iomega[2][j]*_omega2[3][k]);
				//omega^2 appears when _omega2 enter once and hess enter twice, -1} or {_iomega enters twice and hess enters once, (i)^2=-1}
				a2-=_e(i,j,k)*(_omega2[1][i]*_hess[2][j]*_hess[3][k]+_hess[1][i]*_omega2[2][j]*_hess[3][k]+_hess[1][i]*_hess[2][j]*_omega2[3][k]+_iomega[1][i]*_iomega[2][j]*_hess[3][k]+_iomega[1][i]*_hess[2][j]*_iomega[3][k]+_hess[1][i]*_iomega[2][j]*_iomega[3][k]);
				//omega^0 appears only when _hess enters thrice
				a0+=_e(i,j,k)*_hess[1][i]*_hess[2][j]*_hess[3][k];
				}

	//colving cubic equation
	if(a6==0.)//a6=0 means something weird, skip this point for the sake of god
		printf("Something wrong in eigenfrequencies calculation at H=%.4f\nOne frequency is missing. Skipping this point\n",(float)Hval);
	else
		{
		code=Cubic(omega,a4/a6,a2/a6,a0/a6);
		f[0]=f[1]=f[2]=-1.;
		if(code==3)//three real roots, standard
			for(i=0;i<=3;i++)
				f[i]=omega[i]/(4.*M_PI*M_PI);
		if(code==2)//two real roots, degenerate case
			{
			f[0]=omega[0]/(4.*M_PI*M_PI);
			f[1]=f[2]=omega[1]/(4.*M_PI*M_PI);
			}
		if(code==1)
			printf("Something wrong in eigenfrequencies calculation at H=%.4f\nImaginary eigenfrequencies. Setting frequencies to -1 \n",(float)Hval);

		for(i=0;i<3;i++)
			{
			if(f[i]>=0.) f[i]=sqrt(f[i]);//normal
			if(f[i]<0&&f[i]>_cutoff) f[i]=0.;//almost zero, allow for numeric error
			if(f[i]<=_cutoff) f[i]=-1.;
			}
		//sort roots
		for(i=0;i<2;i++)
			for(j=i+1;j<3;j++)
				if(f[i]>f[j])
					{
					sorttmp=f[i];
					f[i]=f[j];
					f[j]=sorttmp;
					}
		}
	handle=fopen(filename,"a");
	fprintf(handle,"%.6f\t%.6g\t%.6g\t%.6g\n",(float)Hval,(float)f[0],(float)f[1],(float)f[2]);
	fclose(handle);

	for(i=1;i<=3;i++)
		omegaout[i]=f[i-1]*2.*M_PI;
	}

void save_eigenvecs(char filename[], double theta, double phi, double psi, double omegain[])
	{
	//complex eigevectors calculation from
	//Mx=0 equation
	//M[i][j]=-omega^2*_omega2[i][j]+i*omega*_iomega[i][j]+_hess[i][j]

	double _hess[4][4],_omega2[4][4],_iomega[4][4];
	float eval[7];
	double phi_re[4][4],phi_im[4][4];
	int i,j,k,alpha,beta,code;
	FILE * handle;
	float **symmatrix,**vmatrix;
	vector3d l[4],dldphi[4][4],H,m_re[4],m_im[4],newvec;
	double maxmod, modm, modtmp, cosphi,sinphi;
	int maxcomp;
	double mpar[4],mper[4],newval;

	H=Hval*dirH;

	#define zerotolerance 1e-4

	//preset m and phi-vectors to zero, 
	//get l-vectors and their derivatives 
	memset(phi_re,0,sizeof(phi_re));
	memset(phi_im,0,sizeof(phi_im));
	for(i=1;i<=3;i++)
		{
		m_re[i]=m_im[i]=makevector(0.,0.,0.);
		l[i]=l_vec(i,theta,phi,psi,&code);
		for(alpha=1;alpha<=3;alpha++)
			dldphi[i][alpha]=dl_dphi(alpha,l[i],&code);
		}

	#define MSIZE 6
	//Numerical Recipes objects identification
	symmatrix=matrix(1,MSIZE,1,MSIZE);
	vmatrix=matrix(1,MSIZE,1,MSIZE);


	//calculate phi-vectors
	for(i=1;i<=3;i++)
		{
		memset(eval,0,sizeof(eval));

		if(omegain[i]<zerotolerance)	//zero frequency means no oscillating magnetization ==> phi=0 as preset;
			continue;					//OR some problem appeared during omega calculations ==> leave phi=0 as preset

		//calculate hessian matrix
		Hessian(_hess,theta,phi,psi,&code);
		//calculate _omega2 and _iomega matrices
		omega_matrices(_omega2,_iomega,theta,phi,psi,&code);

		//fill symmetric matrix
		//to reduce complex problem to real problem
		//see Numerical Recipies 11.4
		for(j=1;j<=3;j++)
			for(k=1;k<=3;k++)
				{
				symmatrix[j][k]=symmatrix[j+3][k+3]=(float)(-omegain[i]*omegain[i]*_omega2[j][k]+_hess[j][k]);
				symmatrix[j+3][k]=(float)(omegain[i]*_iomega[j][k]);
				symmatrix[j][k+3]=-symmatrix[j+3][k];
				}

		//call to eigenvectors routine
		jacobi(symmatrix,MSIZE,eval,vmatrix,&code);

		//find zero eigenvalue and deduce its eigenvector
		//this is rotation vector components, they have to be recalculated to
		//oscillating magnetization later
		for(j=1;j<=6;j++)
			if(fabs(eval[j])<zerotolerance)
				{
				for(k=1;k<=3;k++)
					{
					phi_re[i][k]=(double)vmatrix[k][j];
					phi_im[i][k]=(double)vmatrix[k+3][j];
					}
				break;
				}
		}
		
	
	//calculate m-vector
	for(k=1;k<=3;k++)
		{
		for(i=1;i<=3;i++)
			for(beta=1;beta<=3;beta++)
				{
				m_re[k]=m_re[k]-phi_re[k][beta]*gamma*gamma*I[i]*scalar(dldphi[i][beta],H)*l[i];
				m_re[k]=m_re[k]-phi_re[k][beta]*gamma*gamma*I[i]*scalar(l[i],H)*dldphi[i][beta];
				m_re[k]=m_re[k]-phi_im[k][beta]*gamma*omegain[k]*I[i]*(dldphi[i][beta]*l[i]);

				m_im[k]=m_im[k]+phi_re[k][beta]*gamma*omegain[k]*I[i]*(dldphi[i][beta]*l[i]);
				m_im[k]=m_im[k]-phi_im[k][beta]*gamma*gamma*I[i]*scalar(dldphi[i][beta],H)*l[i];
				m_im[k]=m_im[k]-phi_im[k][beta]*gamma*gamma*I[i]*scalar(l[i],H)*dldphi[i][beta];
				}
		modm=sqrt(0.5)*sqrt(scalar(m_re[k],m_re[k])+scalar(m_im[k],m_im[k]));
		mpar[k]=sqrt(0.5)*sqrt(scalar(m_re[k],dirH)*scalar(m_re[k],dirH)+scalar(m_im[k],dirH)*scalar(m_im[k],dirH));
		mper[k]=sqrt(modm*modm-mpar[k]*mpar[k]);
		if(modm!=0.)
			{
			//norm mpar and mper
			mpar[k]=mpar[k]/modm;
			mper[k]=mper[k]/modm;	
			
			//compensate phase of the largest m-vector component for convenience
			//and norm m-vector 
			maxmod=m_re[k].x*m_re[k].x+m_im[k].x*m_im[k].x;
			maxcomp=1;
			for(j=2;j<=3;j++)
				{
				modtmp=component(m_re[k],j)*component(m_re[k],j)+component(m_im[k],j)*component(m_im[k],j);
				if(modtmp>maxmod)
					{maxmod=modtmp;maxcomp=j;}
				}
			cosphi=component(m_re[k],maxcomp)/sqrt(maxmod);
			sinphi=component(m_im[k],maxcomp)/sqrt(maxmod);
		
			newvec=(1./(sqrt(2.)*modm))*(cosphi*m_re[k]+sinphi*m_im[k]);
			m_im[k]=(1./(sqrt(2.)*modm))*(cosphi*m_im[k]-sinphi*m_re[k]);
			m_re[k]=newvec;
			
			//Finally, set strictly to zero all components 
			//which are less than zerotolerance 
			//to improve readability of output file
			if(fabs(mpar[k])<zerotolerance)
				mpar[k]=0.;
			if(fabs(mper[k])<zerotolerance)
				mper[k]=0.;
			for(beta=1;beta<=3;beta++)
				{
				if(fabs(component(m_re[k],beta))<zerotolerance)
					m_re[k]=set_component(m_re[k],beta,0.);
				if(fabs(component(m_im[k],beta))<zerotolerance)
					m_im[k]=set_component(m_im[k],beta,0.);
				}	
			}
		}

	
	//saving
	handle=fopen(filename,"a");
	fprintf(handle,"%.4f\t",(float)Hval);
	//average magnetization projection
	for(i=1;i<=3;i++)
		fprintf(handle,"%.4g\t%.4g\t",(float)mper[i],(float)mpar[i]);
	//complex magnetization vectors
	for(i=1;i<=3;i++)
		{
		for(j=1;j<=3;j++)
			fprintf(handle,"[%.3g+i*(%.3g)];",(float)(component(m_re[i],j)),(float)(component(m_im[i],j)));
		fprintf(handle,"\t");
		}
	fprintf(handle,"\n");
	fclose(handle);

	//clear Numerical Recipes objects' allocations
	free_matrix(symmatrix,1,MSIZE,1,MSIZE);
	free_matrix(vmatrix,1,MSIZE,1,MSIZE);
	}

#undef _cutoff
#undef MSIZE
#undef zerotolerance
